import base64
import json
from unittest import mock
from bpeasy.tokenizer import BPEasyTokenizer


def test_initialization():
    vocab = {b"hello": 1, b"world": 2}
    tokenizer = BPEasyTokenizer(vocab=vocab)
    assert tokenizer.vocab == vocab
    assert tokenizer.name == "bpeasy"
    assert len(tokenizer.special_tokens) == 0
    assert len(tokenizer) == 2


def test_encode_decode():
    vocab = {b"hello": 1, b" world": 2}
    tokenizer = BPEasyTokenizer(vocab=vocab)
    encoded = tokenizer.encode("hello world", allowed_special="all")
    assert encoded == [1, 2]
    decoded = tokenizer.decode(encoded)
    assert decoded == "hello world"


def test_save_and_load():
    vocab = {b"hello": 1, b" world": 2}
    tokenizer = BPEasyTokenizer(vocab=vocab)

    # Test saving
    with mock.patch("builtins.open", mock.mock_open()) as mock_file:
        tokenizer.save("dummy_path.json")
        mock_file.assert_called_once_with("dummy_path.json", "w")

    # Prepare dummy file content for loading
    dummy_file_content = json.dumps(
        {
            "name": "bpeasy",
            "vocab": {
                base64.b64encode(key).decode("utf-8"): value
                for key, value in vocab.items()
            },
            "regex_pattern": tokenizer.regex_pattern,
            "special_tokens": tokenizer.special_tokens,
        }
    )

    # Test loading
    with mock.patch(
        "builtins.open", mock.mock_open(read_data=dummy_file_content)
    ) as mock_file:
        loaded_tokenizer = BPEasyTokenizer.from_file("dummy_path.json")
        assert loaded_tokenizer.vocab == vocab


@mock.patch("builtins.open", new_callable=mock.mock_open)
@mock.patch("json.dump")
def test_conversion_to_huggingface(mock_json_dump, mock_open):
    vocab = {
        b"h": 0,
        b"e": 1,
        b"l": 2,
        b"o": 3,
        b" ": 4,
        b"w": 5,
        b"r": 6,
        b"d": 7,
        b"he": 8,
        b"ll": 9,
        b"llo": 10,
        b"hello": 11,
        b"wo": 12,
        b"wor": 13,
        b"ld": 14,
        b"world": 15,
        b" world": 16,
    }
    tokenizer = BPEasyTokenizer(vocab=vocab)
    tokenizer.export_to_huggingface_format("dummy_path.json")
    mock_open.assert_called_once_with("dummy_path.json", "w", encoding="utf-8")
    mock_json_dump.assert_called_once()
    args, _ = mock_json_dump.call_args
    assert args[0]["model"]["type"] == "BPE"


@mock.patch("builtins.open", new_callable=mock.mock_open)
@mock.patch("json.dump")
def test_conversion_to_huggingface_with_special_tokens(mock_json_dump, mock_open):
    vocab = {
        b"h": 0,
        b"e": 1,
        b"l": 2,
        b"o": 3,
        b" ": 4,
        b"w": 5,
        b"r": 6,
        b"d": 7,
        b"he": 8,
        b"ll": 9,
        b"llo": 10,
        b"hello": 11,
        b"wo": 12,
        b"wor": 13,
        b"ld": 14,
        b"world": 15,
        b" world": 16,
    }
    tokenizer = BPEasyTokenizer(vocab=vocab, special_tokens=["<|special-0|>", "<pad>"])
    tokenizer.export_to_huggingface_format("dummy_path.json")
    mock_open.assert_called_once_with("dummy_path.json", "w", encoding="utf-8")
    mock_json_dump.assert_called_once()
    args, _ = mock_json_dump.call_args
    assert args[0]["model"]["type"] == "BPE"



from importlib.metadata import version

from .bpeasy import train_bpe

__version__ = version("bpeasy")


__all__ = [
    "save_vocab_to_tiktoken",
    "train_bpe",
    "__version__",
]


def save_vocab_to_tiktoken(
    vocab: dict[bytes, int],
    out_path: str,
    special_tokens: list[str] = [],
    fill_to_nearest_multiple_of_eight: bool = False,
) -> None:
    """
    Export vocab to tiktoken txt format - use this if you want to use tiktoken library directly
    Note: you will need to handle special tokens and regex yourself
    """
    import base64

    sorted_vocab = sorted(list(vocab.items()), key=lambda x: x[1])
    for special_token in special_tokens:
        sorted_vocab.append((special_token.encode("utf-8"), len(sorted_vocab)))

    if fill_to_nearest_multiple_of_eight:
        while len(sorted_vocab) % 8 != 0:
            sorted_vocab.append(
                (f"<|special-{len(sorted_vocab)}|>".encode("utf-8"), len(sorted_vocab))
            )

    with open(out_path, "wb") as f:
        for token, rank in sorted_vocab:
            # encode token to base64 and write to file with rank separated by a space
            f.write(base64.b64encode(token) + b" " + str(rank).encode("utf-8") + b"\n")



import dataclasses
import glob
import json
import logging
import sys
import time
from pathlib import Path

import tokenizers
from tokenizers import Regex, Tokenizer, decoders, pre_tokenizers
from tokenizers.models import BPE
from tokenizers.trainers import BpeTrainer
from tqdm import tqdm

import bpeasy
from bpeasy.tokenizer import BPEasyTokenizer

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


@dataclasses.dataclass
class TrainBPETokenizerArgs:
    dataset: str = "./benchmarks/data"
    vocab_size: int = 32_000
    max_sentencepiece_length: int = 128
    regex_pattern: str = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""

    def __post_init__(self):
        checkpoint_dir = Path(self.dataset)
        assert checkpoint_dir.is_dir(), checkpoint_dir


def jsonl_content_iterator(
    args: TrainBPETokenizerArgs,
):
    """
    Iterates over a jsonl file and yields the content of each line
    Tracks the number of characters yielded and stops when the limit is reached
    This is ripe for optimisation if you want to mess with more fine-grained
    character limits (eg. more Python than Java)
    """
    file_path = args.dataset
    chunk_num, character_count = 0, 0
    chunks = glob.glob(f"{file_path}/*.jsonl")

    while chunk_num < len(chunks):
        file_name = chunks[chunk_num]
        with open(file_name, "r", encoding="utf-8") as f:
            for line in f:
                obj = json.loads(line)
                text = obj["text"]
                text_character_count = len(text)
                character_count += text_character_count
                yield text
        chunk_num += 1


def train_huggingface(args: TrainBPETokenizerArgs):
    # should be at least 0.14.0 to train with char limit
    assert tokenizers.__version__ >= "0.14.0"
    tokenizer = Tokenizer(BPE(byte_fallback=True))
    trainer = BpeTrainer(
        vocab_size=args.vocab_size,
        special_tokens=[f"<0x{i:02X}>" for i in range(256)],  # seed sm vocab
        max_token_length=args.max_sentencepiece_length,
        show_progress=False,
    )
    gpt_regex = Regex(args.regex_pattern)

    split_pre_tokenizer = pre_tokenizers.Split(
        gpt_regex, behavior="isolated", invert=False
    )
    byte_pre_tokenizer = pre_tokenizers.ByteLevel(
        add_prefix_space=False, use_regex=False
    )
    tokenizer.pre_tokenizer = pre_tokenizers.Sequence(
        [split_pre_tokenizer, byte_pre_tokenizer]
    )
    # Use ByteLevel Decoder
    tokenizer.decoder = decoders.Sequence(
        [decoders.ByteLevel(), decoders.ByteFallback()]
    )
    iterator = jsonl_content_iterator(args)
    # training the tokenizer
    tokenizer.train_from_iterator(iterator, trainer)

    return tokenizer


def train_bpeasy(args: TrainBPETokenizerArgs):
    # Use ByteLevel Decoder
    iterator = jsonl_content_iterator(args)
    # training the tokenizer
    vocab = bpeasy.train_bpe(
        iterator,
        args.regex_pattern,
        args.max_sentencepiece_length,
        args.vocab_size,
    )

    return BPEasyTokenizer(
        vocab,
        args.regex_pattern,
        special_tokens=[],
        fill_to_nearest_multiple_of_eight=False,
    )


def encode(tokenizer, args) -> float:
    iterator = jsonl_content_iterator(args)
    lengths = []
    num_bytes = 0
    for text in iterator:
        num_bytes += len(text.encode("utf-8"))
        encoded = tokenizer.encode(text)
        lengths.append(len(encoded))
    return num_bytes / sum(lengths)


def get_mean_std_dev(times: list[float]) -> tuple[float, float]:
    avg_time = sum(times) / len(times)
    std_dev = sum([(t - avg_time) ** 2 for t in times])
    return avg_time, std_dev


if __name__ == "__main__":
    args = TrainBPETokenizerArgs()

    times_train_huggingface = []
    times_encode_huggingface = []
    times_train_bpeasy = []
    times_encode_bpeasy = []
    byte_per_token_bpeasy = []

    for v in tqdm(range(5000, 100_000, 5000)):
        args.vocab_size = v

        time_now = time.time()
        tokenizer = train_huggingface(args)
        times_train_huggingface.append(time.time() - time_now)

        time_now = time.time()
        byte_per_token_hf = encode(tokenizer, args)
        times_encode_huggingface.append(time.time() - time_now)

        time_now = time.time()
        tokenizer = train_bpeasy(args)
        times_train_bpeasy.append(time.time() - time_now)

        time_now = time.time()
        byte_per_token_bpeasy.append(encode(tokenizer, args) / byte_per_token_hf)
        times_encode_bpeasy.append(time.time() - time_now)

    m_hf, std_hf = get_mean_std_dev(times_train_huggingface)
    m_bpeasy, std_bpeasy = get_mean_std_dev(times_train_bpeasy)

    print(f"huggingface train time {m_hf} +/- {std_hf}")
    print(f"bpeasy train time {m_bpeasy} +/- {std_bpeasy}")

    m_hf, std_hf = get_mean_std_dev(times_encode_huggingface)
    m_bpeasy, std_bpeasy = get_mean_std_dev(times_encode_bpeasy)

    print(f"huggingface encode time {m_hf} +/- {std_hf}")
    print(f"bpeasy encode time {m_bpeasy} +/- {std_bpeasy}")

    m_bpeasy, std_bpeasy = get_mean_std_dev(byte_per_token_bpeasy)
    print(f"bpeasy bytes/token vs hf: {m_bpeasy} +/- {std_bpeasy}")



import tiktoken

from typing import Optional
from functools import lru_cache
import json


# Adapted from https://github.com/openai/tiktoken/issues/60#issuecomment-1499977960
def bpe(
    mergeable_ranks: dict[bytes, int], token: bytes, max_rank: Optional[int] = None
) -> list[bytes]:
    parts = [bytes([b]) for b in token]
    while True:
        min_idx = None
        min_rank = None
        for i, pair in enumerate(zip(parts[:-1], parts[1:])):
            rank = mergeable_ranks.get(pair[0] + pair[1])
            if rank is not None and (min_rank is None or rank < min_rank):
                min_idx = i
                min_rank = rank
        if min_rank is None or (max_rank is not None and min_rank >= max_rank):
            break
        assert min_idx is not None
        parts = (
            parts[:min_idx]
            + [parts[min_idx] + parts[min_idx + 1]]
            + parts[min_idx + 2 :]
        )
    return parts


# Source taken from https://github.com/huggingface/transformers/blob/73de5108e172112bc620cfc0ceebfd27730dba11/src/transformers/models/gpt2/tokenization_gpt2.py#L63
@lru_cache()
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a mapping to unicode strings. We specifically avoids mapping to whitespace/control
    characters the bpe code barfs on.

    The reversible bpe codes work on unicode strings. This means you need a large # of unicode characters in your vocab
    if you want to avoid UNKs. When you're at something like a 10B token dataset you end up needing around 5K for
    decent coverage. This is a significant percentage of your normal, say, 32K bpe vocab. To avoid that, we want lookup
    tables between utf-8 bytes and unicode strings.
    """
    bs = (
        list(range(ord("!"), ord("~") + 1))
        + list(range(ord("¡"), ord("¬") + 1))
        + list(range(ord("®"), ord("ÿ") + 1))
    )
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


def convert_tiktoken_to_huggingface(
    encoder: tiktoken.Encoding,
    out_path: str,
    regex_pattern: str = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\\r\\n\\p{L}\\p{N}]?\\p{L}+|\\p{N}{1,3}| ?[^\\s\\p{L}\\p{N}]+[\\r\\n]*|\\s*[\\r\\n]+|\\s+(?!\\S)|\\s+""",
):
    byte_encoder = bytes_to_unicode()

    def token_bytes_to_string(b):
        return "".join([byte_encoder[ord(char)] for char in b.decode("latin-1")])

    def generate_vocab_and_merges(encoder):
        mergeable_ranks = encoder._mergeable_ranks

        merges = []
        vocab = {}
        i = 0
        for token, rank in mergeable_ranks.items():
            # Skip special tokens as they are added separately
            if token_bytes_to_string(token) in encoder._special_tokens:
                continue

            vocab[token_bytes_to_string(token)] = rank

            i += 1
            if len(token) == 1:
                continue
            merged = tuple(bpe(mergeable_ranks, token, max_rank=rank))
            assert len(merged) == 2
            merges.append(" ".join(map(token_bytes_to_string, merged)))

        # Also add special tokens
        vocab.update(encoder._special_tokens)

        return vocab, merges

    vocab, merges = generate_vocab_and_merges(encoder)

    added_tokens = [
        {
            "id": id,
            "content": content,
            "single_word": False,
            "lstrip": False,
            "rstrip": False,
            "normalized": False,
            "special": True,
        }
        for content, id in encoder._special_tokens.items()
    ]

    tokenizer_template = {
        "version": "1.0",
        "truncation": None,
        "padding": None,
        "added_tokens": added_tokens,
        "normalizer": None,
        "pre_tokenizer": {
            "type": "Sequence",
            "pretokenizers": [
                {
                    "type": "Split",
                    "pattern": {"Regex": regex_pattern},
                    "behavior": "Removed",
                    "invert": True,
                },
                {
                    "type": "ByteLevel",
                    "add_prefix_space": False,
                    "trim_offsets": True,
                    "use_regex": False,
                },
            ],
        },
        "post_processor": None,
        "decoder": {
            "type": "ByteLevel",
            "add_prefix_space": True,
            "trim_offsets": True,
            "use_regex": True,
        },
        "model": {
            "type": "BPE",
            "dropout": None,
            "unk_token": None,
            "continuing_subword_prefix": "",
            "end_of_word_suffix": "",
            "fuse_unk": False,
            "byte_fallback": False,
            "vocab": vocab,
            "merges": merges,
        },
    }

    with open(
        out_path,
        "w",
        encoding="utf-8",
    ) as fp:
        json.dump(tokenizer_template, fp, indent=2, ensure_ascii=False)



from bpeasy.convert import bpe


def test_bpe_function():
    mergeable_ranks = {b"ab": 0, b"bc": 1, b"cd": 2}
    token = b"abcd"
    result = bpe(mergeable_ranks, token)
    assert result == [
        b"ab",
        b"cd",
    ], "The bpe function did not split the token correctly"
